class EvaluationState:
    EVALUATION_PENDING="EP"
    EVALUATION_RUNNING="ER"
    EVALUATION_SUCCESSFUL="ES"
    EVALUATION_ERROR="EE"
    
class EpisodeState:
    EPISODE_PENDING="eP"
    EPISODE_RUNNING="eR"
    EPISODE_SUCCESSFUL="eS"
    EPISODE_ERROR="eE"
    EPISODE_ABANDONED="eA"

class Commands:
    GET_GAME_FILE="ggf"
    ACTIVATE_STATE_TRACKING="ast"
    COMPUTE_INTERMEDIATE_REWARD="cir"
    STEP="s"
    RESET="r"
    CLOSE="c"



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mtextworld_remote_env` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``textworld_remote_env.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``textworld_remote_env.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import click
import redis

from textworld_remote_env import messages
from textworld_remote_env import state
from textworld_remote_env.message_broker import ServiceMessageBroker

import json
import numpy as np
import os
import timeout_decorator
import time

import textworld
import crowdai_api
import random

import glob

########################################################
# CONSTANTS
########################################################
PER_STEP_TIMEOUT = 10*60 # 10minutes


class TextWorldRemoteEnvEvaluatorService:
    def __init__(   self,
                    game_paths = [],
                    max_steps = 1000,
                    verbose = False):
        self.game_paths = game_paths
        self.max_steps = max_steps

        self.message_broker = ServiceMessageBroker()

        self.current_game = -1
        self.current_game_begin_time = False
        self.current_env = False
        self.evaluation_state = {}
        self.init_evaluation_state()

    def init_evaluation_state(self):
        absolute_game_paths = []
        for _game_path in self.game_paths:
            assert os.path.exists(_game_path), \
                "GamePath : {} does not exist !".format(_game_path)
            absolute_game_paths.append(
                os.path.abspath(_game_path)
            )
        random.shuffle(absolute_game_paths)
        self.game_paths = absolute_game_paths

        self.evaluation_state["state"] = state.EvaluationState.EVALUATION_PENDING
        self.evaluation_state["episodes"] = []
        for _game_path in self.game_paths:
            _episode_object = {}
            _episode_object["state"] = state.EpisodeState.EPISODE_PENDING
            _episode_object["game"] = os.path.basename(_game_path)
            _episode_object["steps"] = 0
            _episode_object["reward"] = 0
            _episode_object["time"] = 0
            _episode_object["begin_time"] = False
            self.evaluation_state["episodes"].append(
                _episode_object
            )
        
        # Sync Evaluation State with the Oracle
        self.message_broker.sync_info_event_with_oracle(
                self.evaluation_state,
                force=True)

    def get_current_episode_oject(self):
        return self.evaluation_state["episodes"][self.current_game]

    def handle_get_game_file(self):
        self.current_game += 1

        """
        First ensure that if any of the previous episodes are pending 
        (meaning, that the agent  abandoned them before they were "done")
        then we mark them as Abandoned
        """
        for _game_idx in range(self.current_game):
            _episode_object = self.evaluation_state["episodes"][_game_idx]
            if _episode_object["state"] != state.EpisodeState.EPISODE_SUCCESSFUL:
                """
                    TODO: Check with TextWorld team what the best policy 
                    should be here.
                """
                #raise Exception('Agent Abandoned an episode before it was "done"')
                _episode_object["state"] = state.EpisodeState.EPISODE_ABANDONED

        if self.current_game >= len(self.game_paths):
            """
                Do due dilligence to mark the end of the evaluation
            """
            self.evaluation_state["state"] = \
                state.EvaluationState.EVALUATION_SUCCESSFUL

            mean_reward = np.mean(
                [
                    _episode["reward"]
                    for _episode in self.evaluation_state["episodes"]
                ]
            )
            mean_time = np.mean(
                [
                    _episode["time"]
                    for _episode in self.evaluation_state["episodes"]
                ]
            )

            self.evaluation_state["score"] = {
                "score" : mean_reward,
                "score_secondary" : mean_time
            }

            print(  "Successful  evaluation state : ", 
                    json.dumps(self.evaluation_state, indent=4))

            # Sync Evaluation State with the Oracle
            self.message_broker.sync_success_event_with_oracle(
                    self.evaluation_state,
                    force=True)

            """
                Return False
            """
            self.message_broker.send_game_file(False)
        else:
            """
                instantiate a new env with the available game_path
                and return the game_path to the client
                (assuming the same game_path is available to the client too)
            """
            if self.current_env:
                self.current_env.close()

            game_file_path = self.game_paths[self.current_game]
            self.current_env = textworld.start(game_file_path)
            self.message_broker.send_game_file(game_file_path)
            self.evaluation_state["state"] = \
                state.EvaluationState.EVALUATION_RUNNING
            _episode_object = self.get_current_episode_oject()
            _episode_object["begin_time"] = time.time()

            # Sync Evaluation State with the Oracle
            self.message_broker.sync_info_event_with_oracle(
                    self.evaluation_state,
                    force=True)


    def handle_activate_state_tracking(self):
        self.current_env.activate_state_tracking()
        self.message_broker.acknowledge_command()

    def handle_compute_intermediate_reward(self):
        self.current_env.compute_intermediate_reward()
        self.message_broker.acknowledge_command()

    def handle_step(self, _event):
        command = _event["payload"]["command"]
        game_state, reward, done = self.current_env.step(command)

        _episode_object = self.get_current_episode_oject()
        _episode_object["steps"] = game_state.nb_moves
        _episode_object["reward"] = game_state.score
        _episode_object["time"] = time.time() - _episode_object["begin_time"]

        if done:
            _episode_object["state"] = state.EpisodeState.EPISODE_SUCCESSFUL

        # Sync Evaluation State with the Oracle
        self.message_broker.sync_info_event_with_oracle(
                self.evaluation_state,
                force=False)

        self.message_broker.acknowledge_command()


    def handle_reset(self):
        if self.current_env.game_running:
            """
                In the evaluation mode, the agent is allowed to call reset 
                only once on a newly instantiated environment
            """
            raise Exception("Attempt to call env.reset on an already running environment.")

        _episode_object = self.get_current_episode_oject()
        _episode_object["state"] = state.EpisodeState.EPISODE_RUNNING

        # Sync Evaluation State with the Oracle
        self.message_broker.sync_info_event_with_oracle(
                self.evaluation_state,
                force=False)

        self.current_env.reset()
        self.message_broker.acknowledge_command()

    def handle_close(self):
        self.current_env.close()
        self.message_broker.acknowledge_command()

    @timeout_decorator.timeout(10*60) #Time out of 10 minutess
    def get_next_command(self):
        """
            Try to get next command from the agent.
            If the agent doesnt respond within 10 minutes, then throw a timeout
        """
        return next(self.message_broker.remote_handler)

    def run_wrapper(self):
        print("Listening for Agent....")
        while True:
            try:
                _event = self.get_next_command()
            except timeout_decorator.timeout_decorator.TimeoutError:
                raise Exception("Vanishing Agent : Agent hasnt communicated with the evaluator for 10minutes. Abadoning evaluation.")

            print(_event)
            if _event["event_type"] == state.Commands.GET_GAME_FILE:
                self.handle_get_game_file()
            elif _event["event_type"] == state.Commands.ACTIVATE_STATE_TRACKING:
                self.handle_activate_state_tracking()
            elif _event["event_type"] == state.Commands.COMPUTE_INTERMEDIATE_REWARD:
                self.handle_compute_intermediate_reward()
            elif _event["event_type"] == state.Commands.STEP:
                self.handle_step(_event)
            elif _event["event_type"] == state.Commands.RESET:
                self.handle_reset()
            elif _event["event_type"] == state.Commands.CLOSE:
                self.handle_close()

        # Sync Evaluation State with the Oracle
        self.message_broker.sync_info_event_with_oracle(
                self.evaluation_state,
                force=True)


    def run(self):
        try:
            self.run_wrapper()
        except Exception as e:
            self.evaluation_state["state"] = state.EvaluationState.EVALUATION_ERROR
            self.evaluation_state["error"] = str(e)
            if self.current_game >= 0:
                self.evaluation_state["episodes"][self.current_game]["state"] = state.EpisodeState.EPISODE_ERROR

            print(  "Current evaluation state : ",
                    json.dumps(self.evaluation_state, indent=4))
            # Sync Evaluation State with the Oracle
            self.message_broker.sync_error_event_with_oracle(
                    self.evaluation_state,
                    force=True)

@click.command()
@click.option('--game_paths_folder', required=True)
def main(game_paths_folder):
    game_paths = glob.glob(
                    os.path.join(game_paths_folder,"*.ulx")
                    )
    print("Game Paths : ", game_paths)
    service = TextWorldRemoteEnvEvaluatorService(
        game_paths = game_paths
    )
    service.run()



class TEXTWORLD_REMOTE_ENV:
    PING = "TEXTWORLD_REMOTE_ENV.PING"
    PONG = "TEXTWORLD_REMOTE_ENV.PONG"
    ENV_CREATE = "TEXTWORLD_REMOTE_ENV.ENV_CREATE"
    ENV_CREATE_RESPONSE = "TEXTWORLD_REMOTE_ENV.ENV_CREATE_RESPONSE"
    ENV_RESET = "TEXTWORLD_REMOTE_ENV.ENV_RESET"
    ENV_RESET_RESPONSE = "TEXTWORLD_REMOTE_ENV.ENV_RESET_RESPONSE"
    ENV_STEP = "TEXTWORLD_REMOTE_ENV.ENV_STEP"
    ENV_STEP_RESPONSE = "TEXTWORLD_REMOTE_ENV.ENV_STEP_RESPONSE"
    ENV_SUBMIT = "TEXTWORLD_REMOTE_ENV.ENV_SUBMIT"
    ENV_SUBMIT_RESPONSE = "TEXTWORLD_REMOTE_ENV.ENV_SUBMIT_RESPONSE"
    ERROR = "TEXTWORLD_REMOTE_ENV.ERROR"



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.


# -*- coding: utf-8 -*-
import os
import re
import sys
import textwrap
import subprocess
from pkg_resources import Requirement, resource_filename

from typing import Mapping, Union, Tuple, List

import numpy as np

from glk import ffi, lib
from io import StringIO

import textworld
from textworld.generator.game import Game, GameProgression
from textworld.logic import Action, State
from textworld.core import GameNotRunningError

from textworld_remote_env.message_broker import ClientMessageBroker

GLULX_PATH = resource_filename(
                Requirement.parse('textworld'), 
                'textworld/thirdparty/glulx/Git-Glulx'
                )

from textworld.envs.glulx.git_glulx_ml import MissingGameInfosError, \
                    StateTrackingIsRequiredError, OraclePolicyIsRequiredError, \
                    _strip_input_prompt_symbol, _strip_i7_event_debug_tags, \
                    _detect_i7_events_debug_tags, GlulxGameState

class RemoteEnv():
    def __init__(self) -> None:
        self.message_broker = ClientMessageBroker()
    def start(self):
        gamefile = self.message_broker.get_game_file()
        if gamefile == False:
            return False
        else:
            return GlulxEnvironmentWrapper(gamefile)

class GlulxEnvironmentWrapper(textworld.Environment):
    """ 
    
    """
    metadata = {'render.modes': ['human', 'ansi', 'text']}

    def __init__(self, gamefile) -> None:
        """ Creates a GitGlulxML from the given gamefile
        """
        super().__init__()

        self.message_broker = ClientMessageBroker()
        
        self._gamefile = gamefile
        self._process = None

        # Load initial state of the game.
        filename, ext = os.path.splitext(gamefile)
        game_json = filename + ".json"

        if not os.path.isfile(game_json):
            raise MissingGameInfosError()

        self._state_tracking = False
        self._compute_intermediate_reward = False
        self.game = Game.load(game_json)
        self.game_state = None

    def activate_state_tracking(self) -> None:
        self.message_broker.activate_state_tracking()
        self._state_tracking = True

    def compute_intermediate_reward(self) -> None:
        self.message_broker.compute_intermediate_reward()
        self._compute_intermediate_reward = True

    def __del__(self) -> None:
        self.close()

    @property
    def game_running(self) -> bool:
        """ Determines if the game is still running. """
        return self._process is not None and self._process.poll() is None

    def step(self, command: str) -> Tuple[GlulxGameState, float, bool]:
        self.message_broker.step(command)
        if not self.game_running:
            raise GameNotRunningError()

        command = command.strip()
        output = self._send(command)
        if output is None:
            raise GameNotRunningError()

        self.game_state = self.game_state.update(command, output)
        self.game_state.has_timeout = not self.game_running
        return self.game_state, self.game_state.score, self.game_state.game_ended

    def _send(self, command: str) -> Union[str, None]:
        if not self.game_running:
            return None

        if len(command) == 0:
            command = " "

        c_command = ffi.new('char[]', command.encode('utf-8'))
        result = lib.communicate(self._names_struct, c_command)
        if result == ffi.NULL:
            self.close()
            return None

        result = ffi.gc(result, lib.free)
        return ffi.string(result).decode('utf-8')

    def reset(self) -> GlulxGameState:
        self.message_broker.reset()
        if self.game_running:
            self.close()

        self._names_struct = ffi.new('struct sock_names*')

        lib.init_glulx(self._names_struct)
        sock_name = ffi.string(self._names_struct.sock_name).decode('utf-8')
        self._process = subprocess.Popen(["%s/git-glulx-ml" % (GLULX_PATH,), self._gamefile, '-g', sock_name, '-q'])
        c_feedback = lib.get_output_nosend(self._names_struct)
        if c_feedback == ffi.NULL:
            self.close()
            raise ValueError("Game failed to start properly: {}.".format(self._gamefile))
        c_feedback = ffi.gc(c_feedback, lib.free)

        start_output = ffi.string(c_feedback).decode('utf-8')

        self.game_state = GlulxGameState(self)
        self.game_state.init(start_output, self.game, self._state_tracking, self._compute_intermediate_reward)

        # TODO: check if the game was compiled in debug mode. You could parse
        #       the output of the following command to check whether debug mode
        #       was used or not (i.e. invalid action not found).
        self._send('actions')  # Turn on debug print for Inform7 action events.
        self._send('restrict commands')  # Restrict Inform7 commands.

        return self.game_state

    def close(self) -> None:
        self.message_broker.close()
        if self.game_running:
            self._process.kill()
            self._process = None

        try:
            lib.cleanup_glulx(self._names_struct)
        except AttributeError:
            pass  # Attempted to kill before reset

    def render(self, mode: str = "human") -> None:
        outfile = StringIO() if mode in ['ansi', "text"] else sys.stdout

        msg = self.game_state.feedback.rstrip() + "\n"
        if self.display_command_during_render and self.game_state.command is not None:
            msg = '> ' + self.game_state.command + "\n" + msg

        # Wrap each paragraph.
        if mode == "human":
            paragraphs = msg.split("\n")
            paragraphs = ["\n".join(textwrap.wrap(paragraph, width=80)) for paragraph in paragraphs]
            msg = "\n".join(paragraphs)

        outfile.write(msg + "\n")

        if mode == "text":
            outfile.seek(0)
            return outfile.read()

        if mode == 'ansi':
            return outfile



#!usr/bin/env python

import crowdai_api
from textworld_remote_env import state
import time

class ClientMessageBroker:
    def __init__(self):
        self.remote_handler = crowdai_api.events.CrowdAIEvents()
    
    def get_blocking_call_response(self, event_type, payload={}):
        response = self.remote_handler.register_event(
            event_type=event_type,
            payload=payload,
            blocking=True
        )
        if response == None:
            raise Exception("Did you set the CROWDAI_IS_GRADING env variable as True ?")
        return response

    def get_game_file(self):
        response = self.get_blocking_call_response(state.Commands.GET_GAME_FILE)
        game_file = response["game_file"]
        print("Received game_file")
        return game_file
    
    def activate_state_tracking(self):
        response = self.get_blocking_call_response(state.Commands.ACTIVATE_STATE_TRACKING)
        assert response["ack"] == True, "Server did not acknowledge a command"
    
    def compute_intermediate_reward(self):
        response = self.get_blocking_call_response(state.Commands.ACTIVATE_STATE_TRACKING)
        assert response["ack"] == True, "Server did not acknowledge a command"
    
    def step(self, command):
        response = self.get_blocking_call_response(
                    state.Commands.STEP,
                    payload={"command" : command}
                    )
        assert response["ack"] == True, "Server did not acknowledge a command"
    
    def reset(self):
        response = self.get_blocking_call_response(state.Commands.RESET)
        assert response["ack"] == True, "Server did not acknowledge a command"

    def close(self):
        response = self.get_blocking_call_response(state.Commands.CLOSE)
        assert response["ack"] == True, "Server did not acknowledge a command"

        
class ServiceMessageBroker:
    def __init__(self):
        self.remote_handler = crowdai_api.events.CrowdAIEvents()
        self.oracle_handler = crowdai_api.events.CrowdAIEvents(with_oracle=True)

        self.last_oracle_update = 0
        self.oracle_update_frequency = 10 # Every 10 seconds

    def send_game_file(self, game_file_path):
        self.remote_handler.send_blocking_call_response({
            "game_file" : game_file_path
        })

    def acknowledge_command(self):
        self.remote_handler.send_blocking_call_response({
            "ack" : True
        })

    def sync_info_event_with_oracle(self, payload={}, force=False):
        if force or time.time() - self.last_oracle_update > self.oracle_update_frequency:
            self.oracle_handler.register_event(
                event_type=self.oracle_handler.CROWDAI_EVENT_INFO,
                payload=payload
            )
            self.last_oracle_update = time.time()

    def sync_success_event_with_oracle(self, payload={}, force=True):
        if force or time.time() - self.last_oracle_update > self.oracle_update_frequency:
            self.oracle_handler.register_event(
                event_type=self.oracle_handler.CROWDAI_EVENT_SUCCESS,
                payload=payload
            )
            self.last_oracle_update = time.time()

    def sync_error_event_with_oracle(self, payload={}, force=True):
        if force or time.time() - self.last_oracle_update > self.oracle_update_frequency:
            self.oracle_handler.register_event(
                event_type=self.oracle_handler.CROWDAI_EVENT_ERROR,
                payload=payload
            )
            self.last_oracle_update = time.time()
